from typing import Mapping, Union
import numpy as np
from torch import optim
import torch
import logging
import anndata
from torch.utils.tensorboard import SummaryWriter
import torch
import torch.autograd as autograd
import torch.nn.functional as F

from scETM.models import BaseCellModel, BatchClassifier
from scETM.logging_utils import log_arguments
from .UnsupervisedTrainer import UnsupervisedTrainer

_logger = logging.getLogger(__name__)


class MMDTrainer(UnsupervisedTrainer):
    """Docstring here (TODO)
    """

    @log_arguments
    def __init__(self,
        model: BaseCellModel,
        adata: anndata.AnnData,
        ckpt_dir: Union[str, None] = None,
        test_ratio: float = 0.,
        data_split_seed: int = 1,
        init_lr: float = 5e-3,
        lr_decay: float = 5e-5,
        batch_size: int = 2000,
        train_instance_name: str = "scETMbatch",
        restore_epoch: int = 0,
        seed: int = -1,
    ) -> None:
        """Docstring here (TODO)
        """

        super().__init__(model,
            adata,
            ckpt_dir=ckpt_dir,
            test_ratio=test_ratio,
            data_split_seed=data_split_seed,
            init_lr=init_lr,
            lr_decay=lr_decay,
            batch_size=batch_size,
            train_instance_name=train_instance_name,
            restore_epoch=restore_epoch,
            seed=seed
        )

    @log_arguments
    def train(self,
        n_epochs: int = 800,
        eval_every: int = 200,
        n_samplers: int = 4,
        kl_warmup_ratio: float = 1/3,
        min_kl_weight: float = 0.,
        max_kl_weight: float = 1e-7,
        eval: bool = True,
        batch_col: str = "batch_indices",
        save_model_ckpt: bool = True,
        record_log_path: Union[str, None] = None,
        writer: Union[None, SummaryWriter] = None,
        eval_result_log_path: Union[str, None] = None,
        eval_kwargs: Union[None, dict] = None,
        mmd_warmup_ratio: float = 1/6,
        min_mmd_weight: float = 0.,
        max_mmd_weight: float = 1,
        **train_kwargs
    ) -> None:
        train_kwargs.update(dict(
            mmd_warmup_ratio=mmd_warmup_ratio,
            min_mmd_weight=min_mmd_weight,
            max_mmd_weight=max_mmd_weight,
        ))
        return super().train(n_epochs=n_epochs,
            eval_every=eval_every,
            n_samplers=n_samplers,
            kl_warmup_ratio=kl_warmup_ratio,
            min_kl_weight=min_kl_weight,
            max_kl_weight=max_kl_weight,
            eval=eval,
            batch_col=batch_col,
            save_model_ckpt=save_model_ckpt,
            record_log_path=record_log_path,
            writer=writer,
            eval_result_log_path=eval_result_log_path,
            eval_kwargs=eval_kwargs,
            **train_kwargs
        )
    
    def do_train_step(self, dataloader, **kwargs) -> Mapping[str, torch.Tensor]:
        """Docstring (TODO)
        """

        # construct hyper_param_dict
        hyper_param_dict = {
            'kl_weight': self._calc_weight(self.epoch, kwargs['n_epochs'], 0, kwargs['kl_warmup_ratio'], kwargs['min_kl_weight'], kwargs['max_kl_weight']),
            'mmd_weight': self._calc_weight(self.epoch, kwargs['n_epochs'], 0, kwargs['mmd_warmup_ratio'], kwargs['min_mmd_weight'], kwargs['max_mmd_weight'])
        }

        # construct data_dict
        data_dict = {k: v.to(self.device) for k, v in next(dataloader).items()}

        # train for one step, record tracked items (e.g. loss)
        self.model.train()
        if hyper_param_dict['mmd_weight'] > 0.:
            def loss_update_callback(loss, fwd_dict, new_record):
                mmd_loss = mmd(fwd_dict[self.model.clustering_input], data_dict['batch_indices'], self.model.n_batches)
                loss += mmd_loss * hyper_param_dict['mmd_weight']
                new_record['mmd_loss'] = mmd_loss.detach().item()
                return loss, new_record
        else:
            loss_update_callback = None
        new_record = self.model.train_step(self.optimizer, data_dict, hyper_param_dict, loss_update_callback)

        return new_record, hyper_param_dict


def partition(data, partitions, num_partitions):
    res = []
    partitions = partitions.flatten()
    for i in range(num_partitions):
        indices = torch.nonzero((partitions == i), as_tuple=False).squeeze(1)
        res += [data[indices]]
    return res


def pairwise_squared_euclidean(x: torch.Tensor, y: torch.Tensor):
    return ((x.unsqueeze(-1) - y.T) ** 2).sum(1)


def gaussian_kernel_matrix(x, y, alphas):
    """Computes multiscale-RBF kernel between x and y.

        Parameters
        ----------
        x: torch.Tensor
            Tensor with shape [batch_size, z_dim].
        y: torch.Tensor
            Tensor with shape [batch_size, z_dim].
        alphas: Tensor

        Returns
        -------
        Returns the computed multiscale-RBF kernel between x and y.
    """

    dist = pairwise_squared_euclidean(x, y).contiguous()
    beta = 1. / (2. * alphas.view(-1, 1))

    s = beta @ dist.view(1, -1)
    return torch.exp(-s).sum(0).view_as(dist)


def mmd_loss_calc(source_features, target_features):
    """Initializes Maximum Mean Discrepancy(MMD) between source_features and target_features.

        - Gretton, Arthur, et al. "A Kernel Two-Sample Test". 2012.

        Parameters
        ----------
        source_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]
        target_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]

        Returns
        -------
        Returns the computed MMD between x and y.
    """
    alphas = [
        1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1, 5, 10, 15, 20, 25, 30, 35, 100,
        1e3, 1e4, 1e5, 1e6
    ]
    alphas = torch.tensor(alphas, device=source_features.device)

    cost = gaussian_kernel_matrix(source_features, source_features, alphas).mean()
    cost += gaussian_kernel_matrix(target_features, target_features, alphas).mean()
    cost -= 2 * gaussian_kernel_matrix(source_features, target_features, alphas).mean()

    return cost


def mmd(
    embedding: torch.Tensor,
    batch_indices: torch.Tensor,
    n_batches: int,
    n_old_batches: Union[int, None] = None
) -> torch.Tensor:
    """Initializes Maximum Mean Discrepancy(MMD) between every different condition.

        Parameters
        ----------
        n_batches: integer
            Number of batches the data contain.
        n_old_batches: integer
            If not 'None', mmd loss is only calculated on #new batches.
        embedding: torch.Tensor
            Torch Tensor of computed latent data.
        c: torch.Tensor
            Torch Tensor of condition labels.

        Returns
        -------
        Returns MMD loss.
    """

    # partition separates y into num_cls subsets w.r.t. their labels c
    emb_partitions = partition(embedding, batch_indices, n_batches)
    loss = 0.
    if n_old_batches is not None:
        for i in range(n_old_batches):
            for j in range(n_old_batches, n_batches):
                if emb_partitions[i].size(0) < 2 or emb_partitions[j].size(0) < 2:
                    continue
                loss += mmd_loss_calc(emb_partitions[i], emb_partitions[j])
    else:
        for i in range(len(emb_partitions)):
            if emb_partitions[i].size(0) < 1:
                continue
            for j in range(i):
                if emb_partitions[j].size(0) < 1:
                    continue
                loss += mmd_loss_calc(emb_partitions[i], emb_partitions[j])

    return loss



from .trainer_utils import set_seed, prepare_for_transfer, train_test_split
from .UnsupervisedTrainer import UnsupervisedTrainer
from .BatchAdversarialTrainer import BatchAdversarialTrainer
from .MMDTrainer import MMDTrainer



from typing import Sequence, Union
import anndata
import numpy as np
from scipy.sparse.base import spmatrix
import torch
from torch import autograd
import logging
from torch.distributions import kl_divergence
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Normal, Independent, NegativeBinomial

from scETM.logging_utils import log_arguments
from .BaseCellModel import BaseCellModel
from .model_utils import (
    get_fully_connected_layers,
    get_kl
)

_logger = logging.getLogger(__name__)


class scVI(BaseCellModel):

    emb_names = ['z', 's']
    clustering_input = 'z'
    max_logsigma = 10
    min_logsigma = -10
    
    @log_arguments
    def __init__(self,
        n_trainable_genes: int,
        n_batches: int,
        n_fixed_genes: int = 0,
        n_topics: int = 50,
        hidden_sizes: Sequence[int] = (128,),
        bn: bool = True,
        dropout_prob: float = 0.1,
        norm_cells: bool = True,
        normed_loss: bool = False,
        reconstruction_loss: str = "nb",
        input_batch_id: bool = False,
        enable_batch_specific_dispersion: bool = True,
        device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    ):
        super().__init__(n_trainable_genes, n_batches, n_fixed_genes, need_batch=n_batches > 1 and (input_batch_id or enable_batch_specific_dispersion), device=device)

        self.n_topics: int = n_topics
        self.hidden_sizes: Sequence[int] = hidden_sizes
        self.bn: bool = bn
        self.dropout_prob: float = dropout_prob
        self.norm_cells: bool = norm_cells
        self.normed_loss: bool = normed_loss
        self.input_batch_id: bool = input_batch_id
        self.reconstruction_loss: str = reconstruction_loss
        self.enable_batch_specific_dispersion = enable_batch_specific_dispersion
        if self.n_batches <= 1:
            _logger.warning(f'n_batches == {self.n_batches}, disabling batch bias')
            self.input_batch_id = False

        self.encoder = get_fully_connected_layers(
            n_trainable_input=self.n_genes + ((self.n_batches - 1) if self.input_batch_id else 0),
            hidden_sizes=hidden_sizes,
            n_trainable_output=self.n_topics * 4,
            bn=bn,
            dropout_prob=dropout_prob
        )
        
        hidden_sizes = list(hidden_sizes).copy()
        hidden_sizes.reverse()
        
        self.decoder = get_fully_connected_layers(
            n_trainable_input=self.n_topics * 2 + ((self.n_batches - 1) if self.input_batch_id else 0),
            hidden_sizes=hidden_sizes,
            n_trainable_output=self.n_genes,
            bn=bn,
            dropout_prob=dropout_prob
        )

        if self.reconstruction_loss == "mse":
            self.recon_batch_clf = get_fully_connected_layers(
                n_trainable_input=self.n_genes,
                hidden_sizes=self.hidden_sizes,
                n_trainable_output=n_batches,
                bn=bn,
                dropout_prob=dropout_prob
            )
        else:
            if enable_batch_specific_dispersion:
                self.px_total_count = nn.Parameter(torch.randn(self.n_batches, self.n_genes))
            else:
                self.px_total_count = nn.Parameter(torch.randn(1, self.n_genes))
            
        self.to(device)

    def decode(self, z, s, data_dict):
        decoder_inputs = [z, s]
        if self.input_batch_id:
            decoder_inputs.append(self._get_batch_indices_oh(data_dict))
        decoder_input = torch.cat(decoder_inputs, dim=-1)
        px_logits = self.decoder(decoder_input)

        if self.reconstruction_loss != "mse":
            if self.enable_batch_specific_dispersion:
                px_total_count = self.px_total_count[data_dict['batch_indices']].clamp(self.min_logsigma, self.max_logsigma).exp()
            else:
                px_total_count = self.px_total_count[torch.zeros(size=(z.size(0),), dtype=torch.long, device=z.device)].clamp(self.min_logsigma, self.max_logsigma).exp()
        else:
            px_total_count = None
        return px_total_count, px_logits

    def forward(self, data_dict, hyper_param_dict=dict(val=True)):
        cells, library_size = data_dict['cells'], data_dict['library_size']
        normed_cells = cells / library_size
        cells_for_loss = normed_cells if self.normed_loss else cells
        input_cells = normed_cells if self.norm_cells else cells
        if self.input_batch_id:
            input_cells = torch.cat((input_cells, self._get_batch_indices_oh(data_dict)), dim=1)

        mu_qz, logsigma_qz, mu_qs, logsigma_qs = self.encoder(input_cells).chunk(4, dim=-1)
        qz = Independent(Normal(
            loc=mu_qz,
            scale=logsigma_qz.clamp(self.min_logsigma, self.max_logsigma).exp()
        ), 1)
        z = qz.rsample()
        qs = Independent(Normal(
            loc=mu_qs,
            scale=logsigma_qs.clamp(self.min_logsigma, self.max_logsigma).exp()
        ), 1)
        s = qs.rsample()

        if not self.training:
            total_count, logits = self.decode(mu_qz, mu_qs, data_dict)
            fwd_dict = dict(
                z=mu_qz,
                s=mu_qs,
                total_count=total_count,
                logits=logits,
                nll = self.get_reconstruction_loss(cells_for_loss, total_count, logits).sum()
            )
            return fwd_dict

        if self.reconstruction_loss == 'mse':
            perm_mask = np.arange(cells.size(0), dtype=np.int64)
            np.random.shuffle(perm_mask)
            perm_mask = torch.LongTensor(perm_mask)
            _, perm_logits = self.decode(z, s[perm_mask, :], data_dict)
            perm_pred_logit = self.recon_batch_clf(perm_logits.softmax(dim=-1))
            perm_ce = F.cross_entropy(perm_pred_logit, data_dict['batch_indices'][perm_mask])


        total_count, logits = self.decode(z, s, data_dict)
        nll = self.get_reconstruction_loss(cells_for_loss, total_count, logits).mean()
        kl_z = get_kl(mu_qz, logsigma_qz).mean()
        kl_s = get_kl(mu_qs, logsigma_qs).mean()
        loss = nll + hyper_param_dict['kl_weight'] * kl_z + hyper_param_dict['kl_weight'] * kl_s
        if self.reconstruction_loss == 'mse':
            loss += 0.2 * perm_ce
            record = dict(loss=loss, nll=nll, kl_z=kl_z, kl_s=kl_s, perm_ce=perm_ce)
        else:
            record = dict(loss=loss, nll=nll, kl_z=kl_z, kl_s=kl_s, perm_ce=perm_ce)

        record = {k: v.detach().item() for k, v in record.items()}

        fwd_dict = dict(
            z=z,
            s=s,
            total_count=total_count,
            logits=logits
        )
        
        return loss, fwd_dict, record

    def sample_x(self, total_count, logits) -> torch.Tensor:
        # Reconstruction Loss
        if self.reconstruction_loss == "nb":
            x = Independent(NegativeBinomial(total_count=total_count, logits=logits), 1).sample()
        else:
            raise NotImplementedError
        return x

    def get_reconstruction_loss(self, x, total_count, logits) -> torch.Tensor:
        """Return the reconstruction loss (for a minibatch)
        """
        # Reconstruction Loss
        if self.reconstruction_loss == "nb":
            reconst_loss = -Independent(NegativeBinomial(total_count=total_count, logits=logits, validate_args=False), 1).log_prob(x).mean()
        elif self.reconstruction_loss == "mse":
            reconst_loss = F.mse_loss(logits.softmax(dim=-1), x, reduction='none').sum(-1).mean()
        else:
            raise NotImplementedError
        return reconst_loss
